/*
 * Copyright (c) 2019, NVIDIA CORPORATION.  All rights reserved.
 *
 * NVIDIA CORPORATION and its licensors retain all intellectual property
 * and proprietary rights in and to this software, related documentation
 * and any modifications thereto.  Any use, reproduction, disclosure or
 * distribution of this software and related documentation without an express
 * license agreement from NVIDIA CORPORATION is strictly prohibited.
 */

#pragma once

#if USE_DX
# include <DXGI.h>

# if USE_DX11
#  include <d3d11.h>
# endif

# if USE_DX12
#  include <d3d12.h>
# endif

#endif // USE_DX

#if USE_VK
	#if WIN32
		#include <windows.h>
	#endif // WIN32
# include <vulkan/vulkan.h>
# include <vulkan/vulkan.hpp>
# include <nvrhi/vulkan.h>
#endif

#include <GLFW/glfw3.h>
#include <GLFW/glfw3native.h>
#include <nvrhi/nvrhi.h>

#include <list>
#include <array>

struct DefaultMessageCallback : public nvrhi::IMessageCallback
{
    static DefaultMessageCallback& GetInstance();

    virtual void message(nvrhi::MessageSeverity severity, const char* messageText, const char* file = nullptr, int line = 0) override;
};

struct DeviceCreationParameters
{
    bool startMaximized = false;
    bool startFullscreen = false;
    int windowPosX = -1;            // -1 means use default placement
    int windowPosY = -1;
    uint32_t backBufferWidth = 1280;
    uint32_t backBufferHeight = 720;
    uint32_t refreshRate = 0;
    uint32_t swapChainBufferCount = 3;
#if USE_VK
    nvrhi::Format swapChainFormat = nvrhi::Format::SBGRA8_UNORM;
#else
    nvrhi::Format swapChainFormat = nvrhi::Format::SRGBA8_UNORM;
#endif
    uint32_t swapChainSampleCount = 1;
    uint32_t swapChainSampleQuality = 0;
    bool enableDebugRuntime = false;
    bool enableNvrhiValidationLayer = false;
    bool vsyncEnabled = false;

    nvrhi::IMessageCallback* messageCallback = &DefaultMessageCallback::GetInstance();

#if USE_DX
    // Adapter to create the device on. Setting this to non-null overrides adapterNameSubstring.
    // If device creation fails on the specified adapter, it will *not* try any other adapters.
    IDXGIAdapter* adapter = nullptr;
#endif

    // For use in the case of multiple adapters; only effective if 'adapter' is null. If this is non-null, device creation will try to match
    // the given string against an adapter name.  If the specified string exists as a sub-string of the
    // adapter name, the device and window will be created on that adapter.  Case sensitive.
    const char *adapterNameSubstring = nullptr;

    // set to true to enable DPI scale factors to be computed per monitor
    // this will keep the on-screen window size in pixels constant
    //
    // if set to false, the DPI scale factors will be constant but the system
    // may scale the contents of the window based on DPI
    //
    // note that the backbuffer size is never updated automatically; if the app
    // wishes to scale up rendering based on DPI, then it must set this to true
    // and respond to DPI scale factor changes by resizing the backbuffer explicitly
    bool enablePerMonitorDPI = false;

#if USE_DX
    DXGI_USAGE swapChainUsage = DXGI_USAGE_SHADER_INPUT | DXGI_USAGE_RENDER_TARGET_OUTPUT;
    D3D_FEATURE_LEVEL featureLevel = D3D_FEATURE_LEVEL_11_1;
#endif
};

class IRenderPass;

class DeviceManager
{
public:
    static DeviceManager *Create(void);
    bool CreateWindowDeviceAndSwapChain(const DeviceCreationParameters& params, const char *windowTitle);

    void AddRenderPassToFront(IRenderPass *pController);
    void AddRenderPassToBack(IRenderPass *pController);
    void RemoveRenderPass(IRenderPass *pController);

    void RunMessageLoop();

    // returns the size of the window in screen coordinates
    void GetWindowDimensions(int& width, int& height);
    // returns the screen coordinate to pixel coordinate scale factor
    void GetDPIScaleInfo(float& x, float& y)
    {
        x = m_DPIScaleFactorX;
        y = m_DPIScaleFactorY;
    }

    inline nvrhi::IMessageCallback* GetMessageCallback() { return m_pMessageCallback; }

protected:
    bool windowVisible = false;

    DeviceCreationParameters m_DeviceParams;
    GLFWwindow *m_Window = nullptr;
    // set to true if running on NV GPU
    bool m_IsNvidia = false;
    std::list<IRenderPass *> m_vRenderPasses;
    // timestamp in seconds for the previous frame
    double m_PreviousFrameTimestamp = 0.0;
    // current DPI scale info (updated when window moves)
    float m_DPIScaleFactorX = 1.f;
    float m_DPIScaleFactorY = 1.f;

    nvrhi::IMessageCallback* m_pMessageCallback = nullptr;

    double m_AverageFrameTime = 0.0;
    double m_AverageTimeUpdateInterval = 0.5;
    double m_FrameTimeSum = 0.0;
    int m_NumberOfAccumulatedFrames = 0;
    
    DeviceManager() = default;

    void BackBufferResizing();
    void BackBufferResized();

    void Animate(double elapsedTime);
    void Render();
    void UpdateAverageFrameTime(double elapsedTime);

    // device-specific methods
    virtual bool CreateDeviceAndSwapChain() = 0;
    virtual void DestroyDeviceAndSwapChain() = 0;
    virtual void ResizeSwapChain() = 0;
    virtual nvrhi::ITexture* GetBackBuffer() = 0;
    virtual uint32_t GetSwapChainIndex() = 0;
    virtual void BeginFrame() = 0;
    virtual void Present() = 0;

public:
    virtual nvrhi::IDevice *GetDevice() = 0;
    virtual const char *GetRendererString() = 0;

    const DeviceCreationParameters& GetDeviceParams();
    double GetAverageFrameTimeSeconds() { return m_AverageFrameTime; }
    double SetFrameTimeUpdateInterval(double seconds) { m_AverageTimeUpdateInterval = seconds; }
    bool IsVsyncEnabled() { return m_DeviceParams.vsyncEnabled; }
    virtual void SetVsyncEnabled(bool enabled) { m_DeviceParams.vsyncEnabled = enabled; }

    // these are public in order to be called from the GLFW callback functions
    void WindowSizeCallback(int width, int height);
    void WindowCloseCallback() { };
    void WindowIconifyCallback(int iconified) { }
    void WindowFocusCallback(int focused) { }
    void WindowRefreshCallback() { }
    void WindowPosCallback(int xpos, int ypos);

    void KeyboardUpdate(int key, int scancode, int action, int mods);
    void MousePosUpdate(double xpos, double ypos);
    void MouseButtonUpdate(int button, int action, int mods);
    void MouseScrollUpdate(double xoffset, double yoffset);

    void Shutdown();
    virtual ~DeviceManager() { }

#if USE_VK
	virtual void* GetVKInstance() = 0;
	virtual void* GetVKPhysicalDevice() = 0;
	virtual void* GetVKDevice() = 0;
	virtual void* GetVKGraphicsQueue() = 0;
	virtual void* GetVKTransferQueue() = 0;
	virtual void* GetVKComputeQueue() = 0;
	virtual int GetVKGraphicsQueueIndex() = 0;
	virtual int GetVKTransferQueueIndex() = 0;
	virtual int GetVKComputeQueueIndex() = 0;
	virtual const char** GetVKInstanceExtensions() = 0;
	virtual size_t GetVKNumInstanceExtensions() = 0;
	virtual const char** GetVKInstanceLayers() = 0;
	virtual size_t GetVKNumInstanceLayers() = 0;
	virtual const char** GetVKDeviceExtensions() = 0;
	virtual size_t GetVKNumDeviceExtensions() = 0;
#endif
};

class IRenderPass
{
private:
    DeviceManager* m_DeviceManager;

public:
    IRenderPass(DeviceManager* deviceManager)
        : m_DeviceManager(deviceManager)
    { }

    virtual void Render(nvrhi::ITexture* renderTarget, uint32_t swapChainIndex) { }
    virtual void Animate(float fElapsedTimeSeconds) { }
    virtual void BackBufferResizing() { }
    virtual void BackBufferResized(const uint32_t width, const uint32_t height, const uint32_t sampleCount) { }

    // all of these pass in GLFW constants as arguments
    // see http://www.glfw.org/docs/latest/input.html
    // return value is true if the event was consumed by this render pass, false if it should be passed on
    virtual bool KeyboardUpdate(int key, int scancode, int action, int mods) { return false; }
    virtual bool MousePosUpdate(double xpos, double ypos) { return false; }
    virtual bool MouseScrollUpdate(double xoffset, double yoffset) { return false; }
    virtual bool MouseButtonUpdate(int button, int action, int mods) { return false; }

    inline DeviceManager* GetDeviceManager() { return m_DeviceManager; }
    inline nvrhi::IDevice* GetDevice() { return m_DeviceManager->GetDevice(); }
};
